// matrix4.cpp: implementation of the matrix4 class.
//
//////////////////////////////////////////////////////////////////////
#include "matrix4.h"
#include "malloc.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

matrix4::matrix4()
{	
	m = (float*) malloc(16 * sizeof(float));
	destroy=true;
	Identity();
}

matrix4::matrix4(const matrix4& mat)
{
	m = (float*) malloc(16 * sizeof(float));
	for (int i=0;i<16;i++)
		m[i] = mat.m[i];
	//memcpy(m,mat.m,16*sizeof(float));
	destroy=true;
}

matrix4::matrix4(float* mat, bool Destroy)
{
	m = mat;
	destroy=Destroy;
	//memcpy(m, mat, 16*sizeof(float));
}

matrix4::~matrix4()
{
	if (destroy)
		free(m);
}

void matrix4::Clear()
{
	for (int i=0;i<16;i++)
		m[i] = 0.0f;
	//memset(m,0,16*sizeof(float));
}

void matrix4::Assign(float* mat)
{
	for (int i=0;i<16;i++)
		mat[i] = m[i];
	//memcpy(mat, m, 16*sizeof(float));
}

matrix4& matrix4::Identity()
{
	Clear();
	int i;
	for (i=0;i<4;i++)
		m[i*4+i]=1.0f;
	return *this;
}

matrix4& matrix4::Translation(const SFVec3f &v)
{
	Identity();
	m[3*4+0]=v.x;
	m[3*4+1]=v.y;
	m[3*4+2]=v.z;
	return *this;
}

matrix4& matrix4::Rotation(const SFVec3f &v)
{
	float ch = cos(v.y);
    float sh = sin(v.y);
    float ca = cos(v.x);
    float sa = sin(v.x);
    float cb = cos(v.z);
    float sb = sin(v.z);

	Identity();

    m[0*4+0] = ch * ca;
    m[0*4+1] = sh*sb - ch*sa*cb;
    m[0*4+2] = ch*sa*sb + sh*cb;
    m[1*4+0] = sa;
    m[1*4+1] = ca*cb;
    m[1*4+2] = -ca*sb;
    m[2*4+0] = -sh*ca;
    m[2*4+1] = sh*sa*cb + ch*sb;
    m[2*4+2] = -sh*sa*sb + ch*cb;
	return *this;
}

SFRotation matrix4::makeQuat(SFRotation q)
{
	if ( q.x == 0 && q.y==0 && q.z==0 )
	{
		q.z = 1.0f;
		q.q = 0.0f;
	}
    float sintd2 = sin(q.q * 0.5f);
    float len = sqrt(q.x*q.x + q.y*q.y + q.z*q.z);
    float f = sintd2 / len;
    q.q = cos(q.q * 0.5f);
    q.x = q.x * f;
    q.y = q.y * f;
    q.z = q.z * f;
	return q;
}

SFRotation matrix4::unMakeQuat(SFRotation q)
{
	double sina2 = sqrt(q.x*q.x + q.y*q.y + q.z*q.z);
    double angle = (2.0 * atan2((float)sina2, (float)q.q));

    if (sina2 >= 1e-8) {
		sina2 = 1.0/sina2;
        q.x *= sina2;
        q.y *= sina2;
        q.z *= sina2;
        q.q = angle;
    } else {
		q.x = 0;
		q.y = 1;
		q.z = 0;
		q.q = 0;
    }
	return q;
}

matrix4& matrix4::Rotation(const SFRotation q)
{
	Clear();

	float sqw = q.q*q.q;
    float sqx = q.x*q.x;
    float sqy = q.y*q.y;
    float sqz = q.z*q.z;
    m[0*4+0] =  sqx - sqy - sqz + sqw;
    m[1*4+1] = -sqx + sqy - sqz + sqw;
    m[2*4+2] = -sqx - sqy + sqz + sqw;
    
    float tmp1 = q.x*q.y;
    float tmp2 = q.z*q.q;
    m[1*4+0] = 2.0f * (tmp1 + tmp2);
    m[0*4+1] = 2.0f * (tmp1 - tmp2);
    
    tmp1 = q.x*q.z;
    tmp2 = q.y*q.q;
    m[2*4+0] = 2.0f * (tmp1 - tmp2);
    m[0*4+2] = 2.0f * (tmp1 + tmp2);
    tmp1 = q.y*q.z;
    tmp2 = q.x*q.q;
    m[2*4+1] = 2.0f * (tmp1 + tmp2);
    m[1*4+2] = 2.0f * (tmp1 - tmp2);
	
	m[3*4+3]=1;
	
	return *this;
}

matrix4& matrix4::Scale(const SFVec3f &v)
{
	Identity();
	m[0 * 4 + 0] = v.x;
	m[1 * 4 + 1] = v.y;
	m[2 * 4 + 2] = v.z;
	return *this;	
}

float matrix4::Determinant()
{
	float value;
	value =
	m[0*4+3] * m[1*4+2] * m[2*4+1] * m[3*4+0] - m[0*4+2] * m[1*4+3] * m[2*4+1] * m[3*4+0] - m[0*4+3] * m[1*4+1] * m[2*4+2] * m[3*4+0] + m[0*4+1] * m[1*4+3] * m[2*4+2] * m[3*4+0]+
	m[0*4+2] * m[1*4+1] * m[2*4+3] * m[3*4+0] - m[0*4+1] * m[1*4+2] * m[2*4+3] * m[3*4+0] - m[0*4+3] * m[1*4+2] * m[2*4+0] * m[3*4+1] + m[0*4+2] * m[1*4+3] * m[2*4+0] * m[3*4+1]+
	m[0*4+3] * m[1*4+0] * m[2*4+2] * m[3*4+1] - m[0*4+0] * m[1*4+3] * m[2*4+2] * m[3*4+1] - m[0*4+2] * m[1*4+0] * m[2*4+3] * m[3*4+1] + m[0*4+0] * m[1*4+2] * m[2*4+3] * m[3*4+1]+
	m[0*4+3] * m[1*4+1] * m[2*4+0] * m[3*4+2] - m[0*4+1] * m[1*4+3] * m[2*4+0] * m[3*4+2] - m[0*4+3] * m[1*4+0] * m[2*4+1] * m[3*4+2] + m[0*4+0] * m[1*4+3] * m[2*4+1] * m[3*4+2]+
	m[0*4+1] * m[1*4+0] * m[2*4+3] * m[3*4+2] - m[0*4+0] * m[1*4+1] * m[2*4+3] * m[3*4+2] - m[0*4+2] * m[1*4+1] * m[2*4+0] * m[3*4+3] + m[0*4+1] * m[1*4+2] * m[2*4+0] * m[3*4+3]+
	m[0*4+2] * m[1*4+0] * m[2*4+1] * m[3*4+3] - m[0*4+0] * m[1*4+2] * m[2*4+1] * m[3*4+3] - m[0*4+1] * m[1*4+0] * m[2*4+2] * m[3*4+3] + m[0*4+0] * m[1*4+1] * m[2*4+2] * m[3*4+3];
	return value;
}

matrix4	matrix4::Invert()
{
	float* rez = (float*) malloc(16 * sizeof(float));
	rez[0*4+0] = m[1*4+2]*m[2*4+3]*m[3*4+1] - m[1*4+3]*m[2*4+2]*m[3*4+1] + m[1*4+3]*m[2*4+1]*m[3*4+2] - m[1*4+1]*m[2*4+3]*m[3*4+2] - m[1*4+2]*m[2*4+1]*m[3*4+3] + m[1*4+1]*m[2*4+2]*m[3*4+3];
	rez[0*4+1] = m[0*4+3]*m[2*4+2]*m[3*4+1] - m[0*4+2]*m[2*4+3]*m[3*4+1] - m[0*4+3]*m[2*4+1]*m[3*4+2] + m[0*4+1]*m[2*4+3]*m[3*4+2] + m[0*4+2]*m[2*4+1]*m[3*4+3] - m[0*4+1]*m[2*4+2]*m[3*4+3];
	rez[0*4+2] = m[0*4+2]*m[1*4+3]*m[3*4+1] - m[0*4+3]*m[1*4+2]*m[3*4+1] + m[0*4+3]*m[1*4+1]*m[3*4+2] - m[0*4+1]*m[1*4+3]*m[3*4+2] - m[0*4+2]*m[1*4+1]*m[3*4+3] + m[0*4+1]*m[1*4+2]*m[3*4+3];
	rez[0*4+3] = m[0*4+3]*m[1*4+2]*m[2*4+1] - m[0*4+2]*m[1*4+3]*m[2*4+1] - m[0*4+3]*m[1*4+1]*m[2*4+2] + m[0*4+1]*m[1*4+3]*m[2*4+2] + m[0*4+2]*m[1*4+1]*m[2*4+3] - m[0*4+1]*m[1*4+2]*m[2*4+3];
	rez[1*4+0] = m[1*4+3]*m[2*4+2]*m[3*4+0] - m[1*4+2]*m[2*4+3]*m[3*4+0] - m[1*4+3]*m[2*4+0]*m[3*4+2] + m[1*4+0]*m[2*4+3]*m[3*4+2] + m[1*4+2]*m[2*4+0]*m[3*4+3] - m[1*4+0]*m[2*4+2]*m[3*4+3];
	rez[1*4+1] = m[0*4+2]*m[2*4+3]*m[3*4+0] - m[0*4+3]*m[2*4+2]*m[3*4+0] + m[0*4+3]*m[2*4+0]*m[3*4+2] - m[0*4+0]*m[2*4+3]*m[3*4+2] - m[0*4+2]*m[2*4+0]*m[3*4+3] + m[0*4+0]*m[2*4+2]*m[3*4+3];
	rez[1*4+2] = m[0*4+3]*m[1*4+2]*m[3*4+0] - m[0*4+2]*m[1*4+3]*m[3*4+0] - m[0*4+3]*m[1*4+0]*m[3*4+2] + m[0*4+0]*m[1*4+3]*m[3*4+2] + m[0*4+2]*m[1*4+0]*m[3*4+3] - m[0*4+0]*m[1*4+2]*m[3*4+3];
	rez[1*4+3] = m[0*4+2]*m[1*4+3]*m[2*4+0] - m[0*4+3]*m[1*4+2]*m[2*4+0] + m[0*4+3]*m[1*4+0]*m[2*4+2] - m[0*4+0]*m[1*4+3]*m[2*4+2] - m[0*4+2]*m[1*4+0]*m[2*4+3] + m[0*4+0]*m[1*4+2]*m[2*4+3];
	rez[2*4+0] = m[1*4+1]*m[2*4+3]*m[3*4+0] - m[1*4+3]*m[2*4+1]*m[3*4+0] + m[1*4+3]*m[2*4+0]*m[3*4+1] - m[1*4+0]*m[2*4+3]*m[3*4+1] - m[1*4+1]*m[2*4+0]*m[3*4+3] + m[1*4+0]*m[2*4+1]*m[3*4+3];
	rez[2*4+1] = m[0*4+3]*m[2*4+1]*m[3*4+0] - m[0*4+1]*m[2*4+3]*m[3*4+0] - m[0*4+3]*m[2*4+0]*m[3*4+1] + m[0*4+0]*m[2*4+3]*m[3*4+1] + m[0*4+1]*m[2*4+0]*m[3*4+3] - m[0*4+0]*m[2*4+1]*m[3*4+3];
	rez[2*4+2] = m[0*4+1]*m[1*4+3]*m[3*4+0] - m[0*4+3]*m[1*4+1]*m[3*4+0] + m[0*4+3]*m[1*4+0]*m[3*4+1] - m[0*4+0]*m[1*4+3]*m[3*4+1] - m[0*4+1]*m[1*4+0]*m[3*4+3] + m[0*4+0]*m[1*4+1]*m[3*4+3];
	rez[2*4+3] = m[0*4+3]*m[1*4+1]*m[2*4+0] - m[0*4+1]*m[1*4+3]*m[2*4+0] - m[0*4+3]*m[1*4+0]*m[2*4+1] + m[0*4+0]*m[1*4+3]*m[2*4+1] + m[0*4+1]*m[1*4+0]*m[2*4+3] - m[0*4+0]*m[1*4+1]*m[2*4+3];
	rez[3*4+0] = m[1*4+2]*m[2*4+1]*m[3*4+0] - m[1*4+1]*m[2*4+2]*m[3*4+0] - m[1*4+2]*m[2*4+0]*m[3*4+1] + m[1*4+0]*m[2*4+2]*m[3*4+1] + m[1*4+1]*m[2*4+0]*m[3*4+2] - m[1*4+0]*m[2*4+1]*m[3*4+2];
	rez[3*4+1] = m[0*4+1]*m[2*4+2]*m[3*4+0] - m[0*4+2]*m[2*4+1]*m[3*4+0] + m[0*4+2]*m[2*4+0]*m[3*4+1] - m[0*4+0]*m[2*4+2]*m[3*4+1] - m[0*4+1]*m[2*4+0]*m[3*4+2] + m[0*4+0]*m[2*4+1]*m[3*4+2];
	rez[3*4+2] = m[0*4+2]*m[1*4+1]*m[3*4+0] - m[0*4+1]*m[1*4+2]*m[3*4+0] - m[0*4+2]*m[1*4+0]*m[3*4+1] + m[0*4+0]*m[1*4+2]*m[3*4+1] + m[0*4+1]*m[1*4+0]*m[3*4+2] - m[0*4+0]*m[1*4+1]*m[3*4+2];
	rez[3*4+3] = m[0*4+1]*m[1*4+2]*m[2*4+0] - m[0*4+2]*m[1*4+1]*m[2*4+0] + m[0*4+2]*m[1*4+0]*m[2*4+1] - m[0*4+0]*m[1*4+2]*m[2*4+1] - m[0*4+1]*m[1*4+0]*m[2*4+2] + m[0*4+0]*m[1*4+1]*m[2*4+2];
	float det=Determinant();
	int i,j;
	for (i=0;i<4;i++)
		for (j=0;j<4;j++)
			rez[i*4+j]/=det;
	matrix4 matr(rez,true);
	return matr;
}

matrix4	matrix4::operator*(matrix4 mat)
{
	//this->DumpToFile("Multiply first",0,0);
	//mat.DumpToFile("Multiply second",0,0);
	int i,j,c;
	float* rez = (float*) malloc(16 * sizeof(float));
	memset( rez, 0, 16*sizeof(float) );
	for (i=0;i<4;i++)
		for (j=0;j<4;j++)
			for (c=0;c<4;c++)
				rez[i*4+j] += m[i*4+c] * mat.m[c*4+j]; //Data(c,i);
	matrix4 matr( rez, true );
	//matr.DumpToFile("Multiply result",0,0);
	return matr;
}

matrix4& matrix4::operator=(const matrix4 mat)
{
	for (int i=0;i<16;i++)
		m[i] = mat.m[i];
	//memcpy(m,mat.m,16*sizeof(float));
	return *this;
}

void matrix4::DumpToFile(const char* message, int boneid, int parentid)
{
	FILE* f;
	f=fopen("c:\\Matrix_log.txt","w");
	fprintf(f,"%s : %d,  %d\n\n",message,boneid,parentid);
	for (int i=0;i<4;i++)
	{
		for (int j=0;j<4;j++)
			fprintf(f,"%7.3f ",m[i*4+j]);
		fprintf(f,"\n");
	}
	fprintf(f,"\n");
	fprintf(f,"------------------------\n\n");
	fclose(f);
}
