// KMeans.cpp: implementation of the KMeans class.
//
//////////////////////////////////////////////////////////////////////

#include "KMeans.h"
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#define MAX(x,y) ((x > y)? x : y)
#define MIN(x,y) ((x < y)? x : y)
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
#include "Console.h"

extern CConsole logger;

KMeans::KMeans(float ** data, int n, int d, int k, int NIter){
	d_ = d;				// the vectors dimension
	n_ = n;				// nbr of input vectors
	k_ = k;				// nbr clusters
	NIter_ = NIter;			// nbr of iteration for the minimization process
	data_ = data;		// the input data


	centroids_ = new float* [k_];// clusters centroides

	for (int h = 0; h < k_; h++) {
		centroids_[h] = new float [d_];
	}

	rmse_ = new float [n_];		// the errors
	partition_ = new int [n_];	// the partition
	tags_ = new int [n_];		// tags
	srand((unsigned int)time(NULL));
}

KMeans::~KMeans()
{
	for (int h = 0; h < k_; h++) {
		delete [] centroids_[h];
	}
	delete [] centroids_;
	delete [] rmse_;
	delete [] partition_;
	delete [] tags_;
}

double KMeans::Dist(float * vect1, float * vect2) {	// The L2 Norm
	double dist = 0.0;
	for(int s = 0; s < d_; s++) {
		dist += pow((double) (vect1[s] - vect2[s]), 2.0);
	}
	return sqrt(dist);
}



// returns the position of the vector which is the farthest from the centroids
double KMeans::GetMaxDist(int & pos, int k) {	
	double maxDist = 0.0;
	double dist = 0.0;
	pos = -1;
	for (int p = 0; p < n_; p++) {
		dist = Dist(data_[p], centroids_[0]);	// distance the first centroid
		for (int y = 1; y < k; y++) {
			dist = MIN(Dist(data_[p], centroids_[y]), dist);
		}
		if ((dist > maxDist) && (tags_[p] == 0)) {
			maxDist = dist;
			pos = p;
		}
//		printf("%i %i %f\n", p, tags_[p], dist);
	}
	return maxDist;
}

// returns the position of the nearest centroid
double KMeans::GetMinDist(int & pos, int p) {	
	double dist = 0.0;
	pos = 0;
	double minDist = Dist(data_[p], centroids_[0]);
	for (int y = 1; y < k_; y++) {
		dist = Dist(data_[p], centroids_[y]);
		if (dist < minDist) {
			minDist = dist;
			pos = y;
		}
	}
	return minDist;
}




void KMeans::Initialize() {
	//all tags to 0
	for(int p = 0; p < n_; p++) {
		tags_[p] = 0;
	}

	int p0 = (int) (rand() % n_);
	for(int s = 0; s < d_; s++) {
		centroids_[0][s] = data_[p0][s];
	}
	tags_[p0] = 1;

	int pos = 0;
	for(int p = 1; p < k_; p++) {
		double maxDist = GetMaxDist(pos, p);
		for(int s = 0; s < d_; s++) {
			centroids_[p][s] = data_[pos][s];
		}
		tags_[pos] = 1;
	}
	//all tags to 0
	for(int p = 0; p < n_; p++) {
		tags_[p] = 0;
	}
}

void KMeans::OutputCentroids(char * fileName) {
	FILE * fileO = fopen(fileName, "w");
	for (int h = 0; h < k_; h++) {
		fprintf(fileO, "%f %f\n", centroids_[h][0], centroids_[h][1]);
	}
	fclose(fileO);
}

void KMeans::UpdatePartition() {

	int pos = -1;
	for(int p = 0; p < n_; p++) {
		rmse_[p] = (float) GetMinDist(pos, p);
		partition_[p] = pos;
	}
}


double KMeans::ComputeError() {

	double err = 0.0;
	for(int p = 0; p < n_; p++) {
		err += rmse_[p];
	}
	return err;
}

void KMeans::UpdateCentroids() {

	// all centroids to zero
	// nuber of pts per cluster to zero
	int * nbr = new int[k_];
	for (int h = 0; h < k_; h++) {
		for(int s = 0; s < d_; s++) {
			centroids_[h][s]= 0.0;
		}
		nbr[h] = 0;
	}

	// recompute barycenters
	for(int p = 0; p < n_; p++) {
		int h = partition_[p];
		for(int s = 0; s < d_; s++) {
			centroids_[h][s]+= data_[p][s];
		}
		nbr[h]++;
	}
	for (int h = 0; h < k_; h++) {
		for(int s = 0; s < d_; s++) {
			centroids_[h][s]/= nbr[h];
		}
	}
	
	delete [] nbr;
}



void KMeans::Partition() {
	Initialize();

	//	OutputCentroids("seeds.txt");
	char fileP[1024];
	char fileS[1024];

	// allocate memory
	float ** centroids_old = new float* [k_];// clusters centroides
	for (int h = 0; h < k_; h++) {
		centroids_old[h] = new float [d_];
		for(int s = 0; s < d_; s++) {
			centroids_old[h][s] = centroids_[h][s];
		}
	}


	double errCentroids = 1000.0;
	double SSE = 0.0;
	double SSE_old = 0.0;
	double SSE_delta = 1000.0;
	double ErrMin = 0.000001;
	int u = 0;
	for (u = 0; (u < NIter_) && ((errCentroids > ErrMin ) || (SSE_delta > ErrMin )) ; u++) {
		UpdatePartition();
		UpdateCentroids();
		//------------------------
		// Compute the err
		errCentroids = 0.0;
		for (int h = 0; h < k_; h++) {
			for(int s = 0; s < d_; s++) {
				errCentroids += (centroids_[h][s]-centroids_old[h][s]) * (centroids_[h][s]-centroids_old[h][s]);
				centroids_old[h][s] = centroids_[h][s];
			}
		}
		errCentroids = sqrt(errCentroids);

		//----------
		SSE = ComputeError();
		SSE_delta = fabs(SSE-SSE_old);
		SSE_old = SSE;
		logger.write_2_log("iter = %i SSE = %f errC = %f SSE_delta = %f\n", u, SSE, errCentroids, SSE_delta);
//		sprintf(fileP, "partition%i.txt", u);
//		OutputPartition(fileP);
//		sprintf(fileS, "seeds%i.txt", u);
//		OutputCentroids(fileS);
	}
	logger.write_2_log("iter = %i SSE %f ErrC %f\n", u, ComputeError(), errCentroids);
//	OutputCentroids("seeds.txt");

	// free memory
	for (int h = 0; h < k_; h++) {
		delete [] centroids_old[h];
	}
	delete [] centroids_old;
}

void KMeans::OutputPartition(char * fileName) {
	FILE * fileO = fopen(fileName, "w");
	for (int h = 0; h < n_; h++) {
		fprintf(fileO, "%i %f %f %f\n", partition_[h], data_[h][0], data_[h][1], rmse_[h]);
	}
	fclose(fileO);
}