/*!
 ************************************************************************
 *  \file
 *     HierarchicalPartitionner.cpp
 *  \brief
 *     HierarchicalPartitionner class.
 *  \author
 *     Copyright (C) 2007  ARTEMIS Department INT/GET, Paris, France.
 *
 *     Khaled MAMOU               <khaled.mamou@int-evry.fr>
 *
 *	   Institut National des Telecommunications		tel.: +33 (0)1 60 76 40 94
 *     9, Rue Charles Fourier,						fax.: +33 (0)1 60 76 43 81
 *     91011 Evry Cedex France
 *
 ************************************************************************
 */
#include "hierarchicalPartitionner.h"
#include "AffineTransform.h"
#include "AffineMotionPredictor.h"
#include "Animation.h"
#include "MotionModel.h"
#include "Graph.h"
#include <math.h>
#include "Console.h"
#include "KMeans.h"

extern CConsole logger;

HierarchicalPartitionner::HierarchicalPartitionner(void)
{
	K_ = 0;
}

HierarchicalPartitionner::~HierarchicalPartitionner(void)
{
}


float HierarchicalPartitionner::PredictionError(std::vector<int> &vertices, Animation * anim, float diagonal) {
	double dist = 0.0;
	AffineTransform at;
	AffineMotionPredictor pred;
	pred.Init(anim, &vertices);
	float coord0[3];
	float coord1[3];
	float coordp[3];
	for(int f = 0; f < anim->GetNKeyCoord(); f++){
		pred.ComputeAffineTransform(f, at);
		for(int v=0; v < (int) vertices.size(); v++) {
			int vertex = vertices[v];
			anim->GetCoord(0, vertex, coord0);
			anim->GetCoord(f, vertex, coord1);
			at.Get(coord0, coordp);
			dist += pow((coord1[0]-coordp[0])/diagonal, 2.0) + 
				    pow((coord1[1]-coordp[1])/diagonal, 2.0) + 
					pow((coord1[2]-coordp[2])/diagonal, 2.0) ;
		}
	}
	return (float) dist;
}

bool HierarchicalPartitionner::RefinePartition(Animation * anim, int k0, float & distGlobal, int minDimCluster) {

	float diagonal = (float) anim->ComputeFirstFrameBBDiagonal();

	ComputeClustersNumber();

	// we fill the clusters by pushing their corresponding vertices
	std::vector<int> * clusters  = new std::vector<int> [K_];
	for (int i = 0; i < (int) partition_.size(); i++) {
		clusters[partition_[i]].push_back(i);
	}
	std::vector<int> toKill;
	int nbrToKill = 0;
	for (int k = 0; k < K_; k++) {
		if ( ((int) clusters[k].size() < minDimCluster) || (k==k0)) {
			toKill.push_back(1);
			nbrToKill++;
		}
		else {
			toKill.push_back(0);
		}
	}

	// we create our motion model
	MotionModel motionModel(K_, anim->GetNCoord(), anim->GetNKeyCoord());
	motionModel.SetWeightFromPartiton(partition_);

	//we fill the affine trtansforms
	for (int k = 0; k < K_; k++) {
		AffineTransform at;
		AffineMotionPredictor pred;
		pred.Init(anim, &(clusters[k]));
		for(int f = 0; f < anim->GetNKeyCoord(); f++){
			pred.ComputeAffineTransform(f, at);
			motionModel.SetAffineTransform(at, f, k);
		}
	}
	
	// we allocate mempory for the errors
	AffineTransform at;
	double dist = 0.0;
	double distGlobalOld = 0.0;
	distGlobal = 0.0;
	double distMin = 0.0;
	int clusterMin = 0;

	float coord0[3];
	float coord1[3];
	float coordp[3];

	bool hasChanged = false;
	int nbrChanged = 0;
	for (int v = 0; v < anim->GetNCoord(); v++) {
//		logger.write_2_log("\n\nvertex %i (cluster = %i)\n", v, partition_[v]);
		int partition_v_old = partition_[v];
		int isFirstCluster = 1;
		for (int k = 0; k < K_; k++) {
			dist = 0.0;
			for(int f = 0; f < anim->GetNKeyCoord(); f++){
				motionModel.GetAffineTransform(at, f, k);
				anim->GetCoord(0, v, coord0);
				anim->GetCoord(f, v, coord1);
				at.Get(coord0, coordp);
				dist += pow((coord1[0]-coordp[0])/diagonal, 2.0) + 
					    pow((coord1[1]-coordp[1])/diagonal, 2.0) + 
						pow((coord1[2]-coordp[2])/diagonal, 2.0) ;
			}
			if (k == partition_v_old) {
				distGlobalOld += dist;
			}

			if ( (( isFirstCluster == 1) || (dist < distMin)) && (toKill[k] == 0)) {
				distMin = dist;
				clusterMin = k;
				isFirstCluster = 0;
			}
		}
		distGlobal += (float) distMin;
		if (partition_[v] != clusterMin) {
//			logger.write_2_log("%i -> %i\n", partition_[v], clusterMin);
			partition_[v] = clusterMin;
			hasChanged = true;
			nbrChanged++;
		}
	}
	
	int oldK = K_;
	UpdatePartition();
	ComputeClustersNumber();
	distGlobal = pow(distGlobal / (anim->GetNCoord() * anim->GetNKeyCoord()), 0.5);
	logger.write_2_log("	(E: %f -> %f) (K: %i -> %i) toKill(%i) Changed(%i)\n", 
		pow(distGlobalOld / (anim->GetNCoord() * anim->GetNKeyCoord()), 0.5), 
		distGlobal , 
		oldK, K_, nbrToKill, nbrChanged);
	
	delete [] clusters;

	return hasChanged;
}
bool HierarchicalPartitionner::SavePartition(char * file) {
	FILE * fPartition = fopen(file, "w");
	for (int vertex = 0; vertex < (int) partition_.size(); vertex++){	
		fprintf(fPartition, "%i\n", partition_[vertex]);
	}
	fclose(fPartition);
	return true;
}
void HierarchicalPartitionner::ComputeClustersNumber() {
	K_ = 0;
	for (int i = 0; i < (int) partition_.size(); i++) {
		K_ = MAX(K_, partition_[i]+1);
	}
}
bool HierarchicalPartitionner::LoadPartition(char * file) {
	FILE * fPartition = fopen(file, "r");
	partition_.clear();
	int cluster;
	while(!feof(fPartition)){
		fscanf(fPartition, "%i", &cluster);
		partition_.push_back(cluster);
	}
	fclose(fPartition);
	partition_.pop_back();
	ComputeClustersNumber();
	return true;
}
void HierarchicalPartitionner::UpdatePartition() {
	std::map<int, int> mapp;

	int cluster = 0;
	for (int i = 0; i < (int) partition_.size(); i++) {
		if (mapp.find(partition_[i]) == mapp.end()) {
			mapp[partition_[i]] = cluster;
			cluster++;
		}
		partition_[i] = mapp[partition_[i]];
	}
}
void HierarchicalPartitionner::CopyPartition(std::vector<int> &partition1, std::vector<int> &partition2) {
	partition1.clear();
	for (int u = 0; u < (int) partition2.size(); u++) {
		partition1.push_back(partition2[u]);
	}
}
void HierarchicalPartitionner::PartitionTrivial(Animation * anim){
	partition_.clear();
	for (int v = 0; v < anim->GetNCoord(); v++){
		partition_.push_back(0);
	}
	ComputeClustersNumber();
}


int HierarchicalPartitionner::Partition(Animation * anim, int ** triangles, int nTriangles, double globalMSE, int minDimCluster) {

	float diagonal = (float) anim->ComputeFirstFrameBBDiagonal();

	// we fill the graph
	Graph graph;

	for (int v = 0; v < anim->GetNCoord(); v++){
		graph.AddVertex();
		graph.vertices_[v].error_ = 0.0;
	}

	int nbr = (int) graph.vertices_.size();

	for (int t = 0; t < nTriangles; t++) {
        graph.AddEdge(triangles[t][0], triangles[t][1]);
        graph.AddEdge(triangles[t][1], triangles[t][2]);
        graph.AddEdge(triangles[t][2], triangles[t][0]);
	}

	// We connect the CCs
	graph.ConnectCCs();

	// Errors for all the edges to zero
	for (gEdgeMap::iterator pose = graph.edges_.begin(); pose != graph.edges_.end(); ++pose) {
		(pose->second).error_ = 0.0;
		(pose->second).tag_ = -1;
	}


	//--- we simplify the stuff
	double totalError = 0.0;
	int ES = (int) graph.edges_.size();
	logger.write_2_log("\nHierarchical motion clustering\n");
	logger.tic();
	printf("0%%");

	while ( (graph.edges_.size() >= 1) && 
		    (totalError < globalMSE) ) {
		printf("\r");
		int progress = 100-(int) ((float) graph.edges_.size() * 100.0) / ES;
		printf("%i%%", progress);
		// We recompute the min error and the distances
		gEdgeMap::iterator poseMin = graph.edges_.begin();
		double minDist = 1.0+(poseMin->second).error_;
		
		for (gEdgeMap::iterator pose = graph.edges_.begin(); pose != graph.edges_.end(); ++pose) {

			// we update the edges costs
			if ((pose->second).tag_ == 1) {  // the edge has changed
				(pose->second).tag_ = -1;
				int v1 = (pose->second).v1_;
				int v2 = (pose->second).v2_;

				// We compute the new distance
				//----------------------------------
				// we fill the vertices for the motion compensation
				std::vector<int> vertices;
				vertices.push_back(v1);
				vertices.push_back(v2);
				for (int i = 0; i < (int) graph.vertices_[v1].ancestors_.size(); i++) {
					vertices.push_back(graph.vertices_[v1].ancestors_[i]);
				}
				for (int i = 0; i < (int) graph.vertices_[v2].ancestors_.size(); i++) {
					vertices.push_back(graph.vertices_[v2].ancestors_[i]);
				}
				// we compute the error induced by the ecol operation
				float distEdge = PredictionError(vertices, anim, diagonal);
				(pose->second).error_ = MAX(0.0, distEdge); 
			}

			if ( (pose->second).error_ < minDist) {
				minDist = (pose->second).error_;
				poseMin = pose;
			}
		}

		
		// We apply the optimal ecol
		int v1 = (poseMin->second).v1_;
		int v2 = (poseMin->second).v2_;
		double eError = (poseMin->second).error_;
		double gError = MAX(0, eError - graph.vertices_[v1].error_ - graph.vertices_[v2].error_); 	
		totalError += gError / (anim->GetNCoord() * anim->GetNKeyCoord());

		graph.EdgeCollapse(v1, v2);

		// recompute vertex error
		graph.vertices_[v1].error_ = eError;
		float vError = (float)eError;
	}


	// we fill the partition
	partition_.clear();
	
	for (int vertex = 0; vertex < anim->GetNCoord(); vertex++){	
		partition_.push_back(-1);
	}
	
	int currentCluster = 0;
	for (gVertexMap::iterator pos = graph.vertices_.begin(); pos != graph.vertices_.end(); ++pos) {
		partition_[pos->first] = currentCluster;
		for (std::vector<int>::iterator pos1 = (pos->second).ancestors_.begin(); pos1 != (pos->second).ancestors_.end(); ++pos1) {
			partition_[*pos1] = currentCluster;
		}
		currentCluster++;
    }
	printf("\r");
	logger.toc();

	// We refine the partition
	logger.write_2_log("Partition refinement\n");
	logger.tic();
	int y = 0;
	float distGlobal = 0.0;
	while ((RefinePartition(anim, -1, distGlobal, minDimCluster)) && (y++ < 100));
	logger.toc();
	ComputeClustersNumber();
	return K_;
}
int HierarchicalPartitionner::PartitionKmeans(Animation * anim, int ** triangles, int nTriangles, int nClusters, int minDimCluster) {

	float diagonal = (float) anim->ComputeFirstFrameBBDiagonal();

	// we fill the graph
	Graph graph;

	// we allocate memory
	float ** data = new float* [anim->GetNCoord()];
	for (int h = 0; h < anim->GetNCoord(); h++) {
		data[h] = new float [anim->GetNKeyCoord()*12];
	}
	

	for (int v = 0; v < anim->GetNCoord(); v++){
		graph.AddVertex();
		graph.vertices_[v].error_ = 0.0;
	}

	int nbr = (int) graph.vertices_.size();

	for (int t = 0; t < nTriangles; t++) {
        graph.AddEdge(triangles[t][0], triangles[t][1]);
        graph.AddEdge(triangles[t][1], triangles[t][2]);
        graph.AddEdge(triangles[t][2], triangles[t][0]);
	}

	logger.write_2_log("\nKmeans Init\n");
	logger.tic();


	for (int vertex = 0; vertex < anim->GetNCoord(); vertex++){
		std::vector<int> vertices;
		vertices.push_back(vertex);
		for(std::vector<int>::iterator ed = graph.vertices_[vertex].edges_.begin(); ed != graph.vertices_[vertex].edges_.end(); ed++) {
			int vertex2 = vertex;
			if (graph.edges_[*ed].v1_ == vertex) vertex2 = graph.edges_[*ed].v2_;
			else vertex2 = graph.edges_[*ed].v1_;
			vertices.push_back(vertex2);
		}
		AffineTransform at;
		AffineMotionPredictor pred;
		pred.Init(anim, &vertices);
		for(int f = 0; f < anim->GetNKeyCoord(); f++){
			pred.ComputeAffineTransform(f, at);
			data[vertex][f*12+ 0]= at.a_;
			data[vertex][f*12+ 1]= at.b_;
			data[vertex][f*12+ 2]= at.c_;
			data[vertex][f*12+ 3]= at.d_;
			data[vertex][f*12+ 4]= at.e_;
			data[vertex][f*12+ 5]= at.f_;
			data[vertex][f*12+ 6]= at.g_;
			data[vertex][f*12+ 7]= at.h_;
			data[vertex][f*12+ 8]= at.i_;
			data[vertex][f*12+ 9]= at.tx_/diagonal;
			data[vertex][f*12+10]= at.ty_/diagonal;
			data[vertex][f*12+11]= at.tz_/diagonal;
		}
	}

	// we apply the kmeans
	KMeans kmeans(data, anim->GetNCoord(), anim->GetNKeyCoord()*12, nClusters, 100);
	kmeans.Partition();


	// we empty memory
	for (int h = 0; h < anim->GetNCoord(); h++) {
		delete [] data[h];
	}
	delete [] data;


// we fill the partition
	partition_.clear();
	
	for (int vertex = 0; vertex < anim->GetNCoord(); vertex++){	
		partition_.push_back(-1);
	}
	int currentCluster = 0;
	for (int pos = 0; pos < anim->GetNCoord(); pos++) {
		partition_[pos] = kmeans.partition_[pos]; // rand() % nClusters 
    }
	ComputeClustersNumber();
	logger.toc();

	// We refine the partition
	logger.write_2_log("Kmeans Partition refinement\n");
	logger.tic();
	int y = 0;
	float distGlobal = 0.0;
	while ((RefinePartition(anim, -1, distGlobal, minDimCluster)) && (y++ < 100));
	logger.toc();
	ComputeClustersNumber();
	return K_;
}
